unit ooda;

{ ---------------------------------------------------------------
  ooda - OpenOffice Direct Access component(s)
  version 0.06 - test release
  Copylefts by Mospan Dmitry
  --------------------------------------------------------------- }

{$mode objfpc}{$H+}

{$DEFINE bookmarks}

interface

uses
  Classes, db, LResources;

type
  TCell = record
    NumTemplates : integer;
    IsFormula : boolean;
    Data : string;
  end;

  TRow = record
    Data : string;
    Cells : array of tCell;
    RangeId : integer;
    LastPos : integer;
  end;

  TRange = record
    StartRow,
    EndRow : integer;
    LastMinPos,
    LastMaxPos : integer;
    Completed : boolean;
    Name : string;
    DataSet : TDataSet;
    {$IFDEF bookmarks}
    Bmrk : TBookMark;
    BmrkOrder : integer;
    {$ENDIF}
  end;

  TRows = array of TRow;
  TRanges = array of TRange;

  TOdsReport = class (TComponent)
  private
    FTemplate : string;
    FOutFileName : string;
    FAutoOpen : boolean;
    FAutoClose : boolean;
    FTmpDir : string;
    m : pointer;
    FSize : integer;
    FRows : TRows;
    FRanges : TRanges;
    FDataContainer : TComponent; //TDataModule;
    FOfficeUpHeader,
    FTableHeader : string;
    FBmrkOrder : integer;
    procedure SetAutoOpen (AValue : boolean);
    procedure SetAutoClose (AValue : boolean);
  protected
    function UnpackTemplate : boolean;
    function ExecOsCmd (Path : string; const Comline : array of ansistring) : integer;
    function ReadTemplate : integer;
    function ParseOdsTemplate : boolean;
    function AddRow (Data : string) : integer;
    function AddCell (Data : string) : integer;
    function AddRange (CurrentPage, Range, RangeName : string) : boolean;
    function GetValueByName (DataString, ParamName : string) : string;
    procedure SetValueByName (var DataString : string;
                             ParamName : string;
                             NewValue : string);
    function GetFieldValue (DataSetName, FieldName : string) : TField;
    function FixDataStr (data : string) : string;
    function GetCellDataFromTemplate (data : string) : string;
    function GetCellData (row, cell : integer) : string;
    procedure GetModFormula (var f : string; row : integer);
    procedure OutData;
    procedure CheckDataContainer;
    procedure PackFile;
    procedure ReleaseMemory (CloseDataSets : boolean);
    {$IFDEF bookmarks}
    procedure RestoreBookmarks;
    {$ENDIF}
  public
    constructor Create (AOwner : TComponent); override;
    procedure Run;
  published
    property Template: string read FTemplate write FTemplate;
    property FileName: string read FOutFileName write FOutFileName;
    property DataContainer: TComponent read FDataContainer write FDataContainer;
    property AutoOpen: boolean read FAutoOpen write SetAutoOpen;
    property AutoClose: boolean read FAutoClose write SetAutoClose;
  end;

procedure Register;

implementation
Uses
{$IFDEF LINUX}
  Unix, BaseUnix,
{$ELSE}
  Windows,
{$ENDIF}
  SysUtils, FileUtil;

procedure TOdsReport.SetAutoOpen (AValue : boolean);
begin
  FAutoOpen := AValue;
end;

procedure TOdsReport.SetAutoClose (AValue : boolean);
begin
  FAutoClose := AValue;
end;

constructor TOdsReport.Create (AOwner : TComponent);
begin
  inherited Create (AOwner);
end;

{$IFDEF LINUX}
function TOdsReport.ExecOsCmd (Path : string; const Comline : array of ansistring) : integer;
var
  pid : cint;
  status : cint;
begin
  pid := fpFork;
  if pid = 0 then begin
    fpexecl(Path, Comline);
    fpExit(127);
  end
  else
  if pid=-1 then begin {Fork failed}
    ExecOsCmd := 127;
  end
  else begin
    fpWait (status);
    ExecOsCmd := WEXITSTATUS (status);
  end;
end;
{$ELSE}
function TOdsReport.ExecOsCmd (Path : string; const Comline : array of ansistring) : integer;
//function Zapusk(cmdline, param : String) : boolean;
var
  si : TSTARTUPINFO;
  pi : TPROCESSINFORMATION;
  param : string;
  i : integer;
begin
  ExecOsCmd := 0;
  param := '';
  for i := 0 to Length (Comline)-1 do
    param := param + ' ' + Comline[i];
  ZeroMemory(@si,sizeof(si));
  si.cb:=SizeOf(si);
  si.dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
  si.wShowWindow := SW_HIDE;
  if CreateProcess (nil, PChar(path + param),
                    nil, nil, true, CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                    nil, nil, si, pi) then begin
    WaitForSingleObject (pi.hProcess,INFINITE);
    CloseHandle (pi.hProcess);
    CloseHandle (pi.hThread);
  end
  else
    result := 127;
end;
{$ENDIF}

function TOdsReport.UnpackTemplate : boolean;
Const
  {$IFDEF LINUX}
  unzip = '/usr/bin/unzip';
  {$ELSE}
  unzip = 'unzip';
  {$ENDIF}
var
  x : integer;
  n : integer;
begin
  n := 0;
  repeat
    FTmpDir := IncludeTrailingPathDelimiter (GetTempDir (true)+'ooda-'+IntToStr(Round(Now))+'-'+IntToStr(n));
    n += 1;
  until not DirectoryExists (FTmpDir);
  x := ExecOsCmd (unzip, ['-o', FTemplate, '-d', FTmpDir]);
  UnpackTemplate := (x = 0);
end;

function TOdsReport.ReadTemplate : integer;
var
  f : file;
  TemplateName : string;
begin
  TemplateName := FTmpDir + 'content.xml';
  if FileExists (TemplateName) then begin
    FSize := FileSize (TemplateName);
    assignfile (f, TemplateName);
    reset (f, 1);
    GetMem (m, FSize);
    BlockRead (f, m^, FSize);
    CloseFile (f);
    ReadTemplate := FSize;
  end
  else
    ReadTemplate := 0;
end;

function TOdsReport.AddRow (Data : string) : integer;
var
  idx : integer;
begin
  idx := Length (FRows);
  SetLength (FRows, idx+1);
  FRows[idx].Data:=Data;
  FRows[idx].RangeId := -1;
  FRows[idx].LastPos := -1;
  AddRow := idx+1;
end;

function TOdsReport.AddCell (Data : string) : integer;
var
  idxrow,
  idxcell : integer;
begin
  idxrow := Length (FRows)-1;
  idxcell := Length (FRows [idxrow].Cells);
  SetLength (FRows [idxrow].Cells, idxcell+1);
  FRows [idxrow].Cells [idxcell].Data := Data;
  if GetValueByName (Data, 'table:formula') = '' then
    FRows [idxrow].Cells [idxcell].IsFormula := false
  else
    FRows [idxrow].Cells [idxcell].IsFormula := true;
  { !!! check for links on other sources !!! }
  FRows [idxrow].Cells [idxcell].NumTemplates := 0;
  AddCell := idxrow;
end;

function TOdsReport.AddRange (CurrentPage, Range, RangeName : string) : boolean;
var
  idx, ps, i : integer;
  tmp : string;
  maxrow : integer;
  a1, a2 : integer;
begin
  ps := Pos (':'+CurrentPage+'.', Range);
  if (Pos (CurrentPage+'.', Range) = 1)
    and (ps > 0) then begin
    idx := Length (FRanges);
    SetLength (FRanges, idx+1);
    RangeName := UpperCase (RangeName);
    FRanges[idx].Name:=RangeName;
    FRanges[idx].LastMinPos:=-1;
    FRanges[idx].LastMaxPos:=-1;
    FRanges[idx].Completed:=false;
    { DataSet }
    FRanges[idx].DataSet := nil;
    {$IFDEF bookmarks}
    FRanges[idx].Bmrk := nil;
    FRanges[idx].BmrkOrder := 0;
    {$ENDIF}
    for i := 0 to FDataContainer.ComponentCount-1 do begin
      if (FDataContainer.Components[i] is TDataSet)
        and (('_'+UpperCase (FDataContainer.Components[i].Name)+'_' = RangeName)
          or ((UpperCase (FDataContainer.Components[i].Name) = RangeName)))then begin
        FRanges[idx].DataSet := FDataContainer.Components[i] as TDataSet;
        {$IFDEF bookmarks}
        if FAutoClose then
          FRanges[idx].Bmrk := nil
        else
          FRanges[idx].Bmrk := FRanges[idx].DataSet.GetBookmark;
        {$ENDIF}
      end;
    end;
    if FRanges[idx].DataSet = nil then
      raise exception.Create('Dataset for range '+RangeName+' not found in template.');
    { StartRow }
    tmp := '';
    for i := ps-1 downto 1 do begin
      if (Range [i] >= '0') and (Range [i] <= '9') then tmp := Range [i] + tmp
      else break;
    end;
    FRanges[idx].StartRow:=StrToInt (tmp)-1;
    { EndRow }
    tmp := '';
    for i := Length (Range) downto ps do begin
      if (Range [i] >= '0') and (Range [i] <= '9') then tmp := Range [i] + tmp
      else break;
    end;
    FRanges[idx].EndRow:=StrToInt (tmp)-1;
    maxrow := length (FRows)-1;
    for i := FRanges[idx].StartRow to FRanges[idx].EndRow do begin
      if i > maxrow then break;
      if FRows [i].RangeId < 0 then
        FRows [i].RangeId:=idx
      else begin
        a1 := FRanges [FRows [i].RangeId].StartRow;
        a2 := FRanges [FRows [i].RangeId].EndRow;
        if (FRanges[idx].StartRow >= a1) and (FRanges[idx].EndRow <= a2) then begin
          FRows [i].RangeId:=idx;
        end;
      end;
    end;
    AddRange := true;
  end
  else AddRange := false;
end;


function TOdsReport.GetValueByName (DataString, ParamName : string) : string;
var
  i, j : integer;
begin
  i := Pos (ParamName+'="', DataString);
  if i > 0 then begin
    i := i + Length (ParamName)+2;
    j := i;
    While (DataString [j] <> '"') and (j <= Length (DataString)) do begin
      j += 1;
    end;
    GetValueByName := Copy (DataString, i, j-i);
  end
  else GetValueByName := '';
end;

procedure TOdsReport.SetValueByName (var DataString : string;
                                    ParamName : string;
                                    NewValue : string);
var
  i, j : integer;
begin
  i := Pos (ParamName+'="', DataString);
  if i > 0 then begin { Param Exists }
    j := i+Length(ParamName)+2;
    While (DataString [j] <> '"') and (j <= Length (DataString)) do begin
      j += 1;
    end;
    Delete (DataString, i, j-i+1);
    Insert (ParamName+'="'+NewValue+'"', DataString, i);
  end
  else begin { New Param }
    i := Pos (' ', DataString);
    Insert (ParamName+'="'+NewValue+'" ', DataString, i+1);
  end;
end;


function TOdsReport.ParseOdsTemplate : boolean;
var
  i, lastpos : integer;
  ch, pch : char;
  tmpstr : string;
  intag : integer;
  OfficeTableStart : integer;
  EndOfRows : boolean;
  StartOfRow : integer;
  startcell : integer;
Const
  tgOfficeUpConstraint = '/office:automatic-styles>';
  tgOfficeTable = 'table:table';
  tgOfficeTableEnd = '/table:table>';
  tgTableRow = 'table:table-row';
begin
  ParseOdsTemplate := true;
  { header of document }
  lastpos := 0;
  SetLength (tmpstr, Length (tgOfficeUpConstraint));
  for i := 0 to fsize-1-Length (tgOfficeUpConstraint) do begin
    ch := PChar (m+i)^;
    if ch = '<' then begin
      move ((m+i+1)^, tmpstr [1], Length (tgOfficeUpConstraint));
      if tmpstr = tgOfficeUpConstraint then begin
        lastpos := i + Length (tgOfficeUpConstraint)+1;
        break;
      end;
    end;
  end;
  SetLength (FOfficeUpHeader, lastpos);
  move (m^, FOfficeUpHeader [1], lastpos);
  // OfficeUpHeader - upper header

  { begin of table:table" }
  OfficeTableStart := -1;
  SetLength (tmpstr, Length (tgOfficeTable));
  for i := lastpos to fsize-1-Length (tgOfficeTable) do begin
    ch := PChar (m+i)^;
    if ch = '<' then begin
      move ((m+i+1)^, tmpstr [1], Length (tgOfficeTable));
      if tmpstr = tgOfficeTable then begin
        OfficeTableStart := i;
        break;
      end;
    end;
  end;
  if OfficeTableStart = -1 then begin
    exit;
  end;

  { разобрать таблицу в массив, выбросить лишнее }
  { ищем "table:table-row" }
  SetLength (tmpStr, Length (tgTableRow));
  StartOfRow := -1;
  for i := OfficeTableStart to fsize-1-Length (tgTableRow) do begin
    ch := PChar (m+i)^;
    if ch = '<' then begin
      move ((m+i+1)^, tmpstr [1], Length (tgTableRow));
      if tmpstr = tgTableRow then begin
        StartOfRow := i;
        break;
      end;
    end;
  end;
  if StartOfRow = -1 then begin
    raise exception.create ('Unknown input format.');
  end;
  SetLength (FTableHeader, i-OfficeTableStart);
  move ((m+OfficeTableStart)^, FTableHeader [1], i-OfficeTableStart);

  { row cycle }
  EndOfRows := false;
  startcell := -1;
  pch := #0;
  While not EndOfRows do begin
    for i := StartOfRow to fsize-1-Length (tgOfficeTableEnd) do begin
      ch := PChar (m+i)^;
      if ch = '<' then begin
        intag := i;
      end
      else if ch = '>' then begin
        SetLength (tmpstr, i-intag+1);
        move ((m+intag)^, tmpstr [1], i-intag+1);
        if tmpstr = '<table:table-row>' then begin
          { create new row }
          AddRow (tmpstr);
        end
        else if tmpstr = '<table:table-row/>' then begin
          { create new row }
          AddRow ('<table:table-row>');
        end
        else if Pos ('<table:table-row ', tmpstr) = 1 then begin
          { create new row (tmpstr) }
          if pch = '/' then begin
            tmpstr := Copy (tmpstr, 1, Length (tmpstr)-2)+'>';
          end;
          AddRow (tmpstr);
        end
        else if tmpstr = '<table:table-cell>' then begin
          startcell := intag;
        end
        else if tmpstr = '<table:table-cell/>' then begin
          { fill empty cell }
          if pch = '/' then begin
            tmpstr := Copy (tmpstr, 1, Length (tmpstr)-2)+'>';
          end;
          AddCell (tmpstr);
        end
        else if Pos ('<table:table-cell ', tmpstr) = 1 then begin
          if pch = '/' then begin
            { fill cell (tmpstr) }
            tmpstr := Copy (tmpstr, 1, Length (tmpstr)-2)+'>';
            AddCell (tmpstr);
          end
          else startcell := intag;
        end
        else if tmpstr = '</table:table-cell>' then begin
          SetLength (tmpstr, i-startcell+1-Length (tmpstr));
          move ((m+startcell)^, tmpstr [1], Length(tmpstr));
          { fill cell (tmpstr) }
          AddCell (tmpstr);
        end
        else if tmpstr = '</table:table>' then begin
          EndOfRows := true;
          break;
        end;
      end;
      pch := ch;
    end;
  end;
  lastpos := i;
  { scan ranges }
  for i := lastpos to fsize-1 do begin
    ch := PChar (m+i)^;
    if ch = '<' then begin
      intag := i;
    end
    else if ch = '>' then begin
      SetLength (tmpstr, i-intag+1);
      move ((m+intag)^, tmpstr [1], i-intag+1);
      if Pos ('<table:database-range ', tmpstr) = 1 then begin
        AddRange (GetValueByName (FTableHeader, 'table:name'),
                  GetValueByName(tmpstr, 'table:target-range-address'),
                  GetValueByName(tmpstr, 'table:name'));
      end
      else if (tmpstr = '</table:database-ranges>')
           or (tmpstr = '</office:spreadsheet>') then begin
        break;
      end;
    end;
  end;
end;

function TOdsReport.GetFieldValue (DataSetName, FieldName : string) : TField;
var
  i : integer;
begin
  GetFieldValue := nil;
  for i := 0 to FDataContainer.ComponentCount-1 do begin
    if (FDataContainer.Components[i] is TDataSet)
      and ('##'+UpperCase (FDataContainer.Components[i].Name) = DataSetName) then begin
      GetFieldValue := (FDataContainer.Components[i] as TDataSet).FieldByName(FieldName);
      break;
    end;
  end;
end;

function TOdsReport.FixDataStr (data : string) : string;
var
  i, Len : integer;
begin
  i := 1;
  Len := Length (data);
  While i < Len do begin
    if data [i] = '&' then begin
      Delete (data, i, 1);
      Insert ('&amp;', data, i);
      Len += 4;
      i += 5;
    end
    else if data [i] = '<' then begin
      Delete (data, i, 1);
      Insert ('&lt;', data, i);
      Len += 3;
      i += 4;
    end
    else if data [i] = '>' then begin
      Delete (data, i, 1);
      Insert ('&gt;', data, i);
      Len += 3;
      i += 4;
    end
    else if data [i] = '"' then begin
      Delete (data, i, 1);
      Insert ('&quot;', data, i);
      Len += 5;
      i += 6;
    end
    else
      i += 1;
  end;
  FixDataStr := data;
end;

function TOdsReport.GetCellDataFromTemplate (data : string) : string;
var
  i, j, jj : integer;
  dataset, fieldname : string;
  newdata : TField;
  tmp : string;
begin
  While Pos ('##', data) > 0 do begin
    i := Pos ('##', data);
    dataset := '';
    for j := i to length (data) do begin
      if data [j] = '.' then break;
      dataset := dataset + data [j];
    end;
    if data [j] = '.' then j += 1;
    fieldname := '';
    for jj := j to length (data)-1 do begin
      if Copy (data, jj, 2) = '##' then break;
      fieldname := fieldname + data [jj];
    end;
    if jj = length (data)-1 then fieldname := '';
    if (fieldname <> '') and (dataset <> '') then begin
      newdata := GetFieldValue (UpperCase (dataset), UpperCase (fieldname));
      if newdata <> nil then begin
        if newdata.DataType = ftString then
          tmp := FixDataStr (newdata.AsString)
        else
          tmp := newdata.AsString;
        Delete (data, i, jj-i+2);
        Insert (tmp, data, i);
        Case newdata.DataType of
          ftString: begin {} end;
          ftSmallint,
          ftInteger,
          ftWord,
          ftBoolean,
          ftFloat,
          ftCurrency,
          ftBCD,
          ftLargeint: begin
            if newdata.IsNull then begin
              SetValueByName (data, 'office:value', '');
            end
            else begin
              SetValueByName (data, 'office:value-type', 'float');
              SetValueByName (data, 'office:value', tmp);
            end;
          end;
          ftDate,
          ftTime,
          ftDateTime: begin
            if newdata.IsNull then begin
              SetValueByName (data, 'office:value', '');
            end
            else begin
              SetValueByName (data, 'office:value-type', 'date');
              tmp [11] := 'T';
              SetValueByName (data, 'office:date-value', tmp);
            end;
          end;
          else begin
            raise exception.Create('Unknown datatype '+IntToStr(Integer (newdata.DataType))
                                  +' for '+dataset+'.'+fieldname+'##');
          end;
        end;
      end
      else begin
        raise exception.create ('Can''t get data from '
                               +Copy(dataset, 3, length(dataset))+'.'+fieldname
                               +#13+'Check your template.');
      end;
    end
    else break;
  end;
  GetCellDataFromTemplate := data;
end;

function TOdsReport.GetCellData (row, cell : integer) : string;
var
  data, f : string;
begin
  data := FRows [row].Cells[cell].Data;
  if FRows [row].Cells[cell].IsFormula then begin
    f := GetValueByName (data, 'table:formula');
    GetModFormula (f, row);
    SetValueByName (data, 'table:formula', f);
  end
  else data := GetCellDataFromTemplate (data);
  GetCellData := data;
end;

procedure TOdsReport.GetModFormula (var f : string; row : integer);
var
  i, j : integer;
  pch, ch : char;
  inq : boolean;
  fstr : string;
  ifstr : integer;
  target : string;
  itarget : integer;
  downlink : boolean;
  d : integer;
//  range_id : integer;
begin
//  target :='9';
//  f := 'of:=SUM([.AC31:.C2345])+[.AMJ1]+[.AMJK123]';
  inq := false;
  pch := #0;
  i := 0;
  fstr := '';
  while i < length (f) do begin
    i += 1;
    ch := f [i];
    if ch = '"' then begin
      inq := not inq;
      fstr := '';
    end
    else if not inq then begin
      if (ch = '.') and ((pch = '[') or (pch = ':')) then begin
        j := i+1;
        while (j < length (f)) and ((f [j] >= 'A') and (f [j] <= 'Z')) and (j-i < 4) do
          j += 1;
        if (f [j] >= '0') and (f [j] <= '9') then begin
          fstr := '';
          while (j < length (f)) and (f [j] >= '0') and (f [j] <= '9') do begin
            fstr := fstr + f [j];
            j += 1;
          end;
          ch := f [j];
          if (ch = ']') or (ch = ':') then begin
            ifstr := StrToInt (fstr);
            itarget := 0;
            if row >= ifstr-1 then begin { uplinks :) }
              if (FRows[row].RangeId <> FRows[ifstr-1].RangeId) then begin
                if FRanges [FRows[ifstr-1].RangeId].Completed then begin
                  if pch = '[' then
                    itarget := FRanges [FRows[ifstr-1].RangeId].LastMinPos
                  else
                    itarget := FRanges [FRows[ifstr-1].RangeId].LastMaxPos;
                 end
                 else begin
                   itarget := FRows[ifstr-1].LastPos;
                 end;
              end
              else begin
                itarget := FRows[ifstr-1].LastPos;
              end;
            end
            else begin { downlinks }
              downlink := true;
              { need to move into init section }
              for d := row to ifstr-1 do
                if FRows[row].RangeId <> FRows[d].RangeId then begin
                  downlink := false;
                  break;
                end;
              if downlink then begin
                itarget := FRows[row].LastPos+ifstr-1-row;
              end;
            end;
            {-------}
            Delete (f, j-length(fstr), length (fstr));
            insert (IntToStr (itarget), f, j-length(fstr));
            j := j - length (fstr) + length (target);
          end;
        end;
        fstr := '';
        i := j;
      end;
    end;
    pch := ch;
  end;
end;

Procedure TOdsReport.OutData;
var
  f : textfile;
  LastPos : integer;
  ReportFileName : string;

  Procedure OutRow  (i : integer);
  var
    c : integer;
  begin
    writeln (f, FRows[i].Data);
    LastPos += 1;
    if FRows[i].LastPos <> -1 then begin
      for c := i+1 to Length (FRows)-1 do begin
        if FRows [c].LastPos = -1 then break
        else FRows [c].LastPos := -1;
      end;
    end;
    FRows[i].LastPos:=LastPos;
    for c := 0 to Length (FRows [i].Cells)-1 do begin
      writeln (f, '  ', GetCellData (i, c));
      writeln (f, '  ', '</table:table-cell>');
    end;
    writeln (f, '</table:table-row>');
  end;



  Function OutRange (rangeid : integer) : integer;
  var
    start, finish : integer;
    i: integer;
    RangeDataSet : TDataSet;
    skiprows : integer;
  begin
    skiprows := 0;
    if rangeid < 0 then begin { Not specified }
      start := 0;
      finish := Length (FRows)-1;
      while start <= finish do begin
        if FRows [start].RangeId >= 0 then begin
          start += OutRange (FRows [start].RangeId);
        end
        else begin
          OutRow (start);
          start += 1;
          skiprows += 1;
        end;
      end;
    end
    else begin
      FRanges [rangeid].Completed:=false;
      FRanges [rangeid].LastMinPos:=-1;
      FRanges [rangeid].LastMaxPos:=-1;
      start := FRanges [rangeid].StartRow;
      finish := FRanges [rangeid].EndRow;
      RangeDataSet := FRanges [rangeid].DataSet;

      if FAutoOpen then begin
        if not RangeDataSet.Active then
          RangeDataSet.Active := true
        else
          RangeDataSet.First;
      end;

      {$IFDEF bookmarks}
      if (RangeDataSet.Active)
      and (FRanges [rangeid].BmrkOrder = 0)
      and (not FAutoClose) then begin
        FBmrkOrder += 1;
        FRanges [rangeid].BmrkOrder := FBmrkOrder;
      end;
      {$ENDIF}

      while not RangeDataSet.EOF do begin
        i := start;
        while i <= finish do begin
          if FRows [i].RangeId <> rangeid then begin
            i += OutRange (FRows [i].RangeId);
          end
          else begin
            if FRanges [rangeid].LastMinPos = -1 then
              FRanges [rangeid].LastMinPos := LastPos + 1;
            FRanges [rangeid].LastMaxPos := LastPos + 1;
            OutRow (i);
            i += 1;
            skiprows += 1;
          end;
        end;
        RangeDataSet.Next;
      end;
      //if FAutoClose then RangeDataSet.Close;
      skiprows := finish-start+1;
      FRanges [rangeid].Completed:=true;
    end;
    OutRange := skiprows;
  end;
begin
  ReportFileName := FTmpDir + 'content.xml';
  AssignFile (f, ReportFileName);
  Rewrite (f);
  Writeln (f, FOfficeUpHeader);
  writeln (f, '<office:body>');
  writeln (f, '<office:spreadsheet>');
  SetValueByName (FTableHeader, 'table:name', 'Report');
  Writeln (f, FTableHeader);
  LastPos := 0;
  OutRange (-1);
  writeln (f, '</table:table>');
  writeln (f, '</office:spreadsheet>');
  writeln (f, '</office:body>');
  writeln (f, '</office:document-content>');
  CloseFile (f);
end;

procedure TOdsReport.CheckDataContainer;
begin
  if FDataContainer = nil then
    if Owner <> nil then
      FDataContainer := Owner;
  if not (FDataContainer is TComponent) then
    raise exception.create ('Invalid DataContainer.');
end;

procedure TOdsReport.PackFile;
var
  exec : integer;
  {$IFDEF WINDOWS}
  CurDir : string;
  {$ENDIF}
begin
  {$IFDEF LINUX}
    exec := ExecOsCmd ('/usr/bin/zip', ['-m', FOutFileName, '-r', '/'+FTmpDir]);
    if exec <> 0 then
      raise exception.Create('zip returned errorcode='+IntTostr(exec));
    RemoveDir (FTmpDir);
  {$ELSE}
    CurDir := GetCurrentDir;
    if Pos ('\', FOutFileName) = 0 then begin
      FOutFileName := IncludeTrailingPathDelimiter (CurDir) + FOutFileName;
    end;
    SetCurrentDir (FTmpDir);
    exec := ExecOsCmd ('zip', ['-m', FOutFileName, '-r', '*']);
    SetCurrentDir (CurDir);
    RemoveDir (FTmpDir);
  {$ENDIF}
  if exec <> 0 then
    raise exception.Create('zip returned errorcode='+IntTostr(exec));
end;

procedure TOdsReport.Run;
var
  SaveDecimalSeparator : char;
  SaveShortDateFormat : string;

begin
  if FTemplate = '' then begin
    raise exception.create ('Template filename not specified.');
  end;
  if FileExists (FTemplate) then begin
    if FOutFileName = '' then begin
      raise exception.create ('Target filename not specified.');
    end;
    ReleaseMemory (false);
    CheckDataContainer;
    if UnpackTemplate then begin
      SaveDecimalSeparator := DecimalSeparator;
      SaveShortDateFormat := ShortDateFormat;
      DecimalSeparator := '.';
      ReadTemplate;
      ParseOdsTemplate;
      { do }
      OutData;
      {$IFDEF bookmarks}
      RestoreBookmarks;
      {$ENDIF}
      ReleaseMemory (true);
      PackFile;
      DecimalSeparator := SaveDecimalSeparator;
      ShortDateFormat := SaveShortDateFormat;
    end
    else
      raise exception.create ('Can''t unpack template. unzip not installed?');
  end
  else
    raise exception.create ('Can''t find template '+FTemplate);
end;

procedure TOdsReport.ReleaseMemory (CloseDataSets : boolean);
var
  r, c : integer;
begin
  { Cells and Rows }
  for r := Length (FRows)-1 downto 0 do begin
    for c := Length (FRows [r].Cells)-1 downto 0 do begin
      SetLength (FRows [r].Cells[c].Data, 0);
    end;
    SetLength (FRows [r].Cells, 0);
    SetLength (FRows [r].Data, 0);
  end;
  SetLength (FRows, 0);
  { Ranges }
  for r := Length (FRanges)-1 downto 0 do begin
    SetLength (FRanges [r].Name, 0);
    if (CloseDataSets) and (FAutoClose) then
      if FRanges [r].Dataset.Active then
         FRanges [r].Dataset.Close;
    {$IFDEF bookmarks}
    if FRanges [r].Bmrk <> nil then begin
      FRanges [r].DataSet.FreeBookMark (FRanges [r].Bmrk);
      FRanges [r].Bmrk := nil;
    end;
    {$ENDIF}
  end;
  SetLength (FRanges, 0);
  {$IFDEF bookmarks}
  FBmrkOrder := 0;
  {$ENDIF}
end;

{$IFDEF bookmarks}
procedure TOdsReport.RestoreBookmarks;
var
  i, j : integer;
begin
  for i := 1 to FBmrkOrder do begin
    for j := 0 to Length (FRanges)-1 do begin
      if FRanges [j].BmrkOrder = i then begin
        try
          if FRanges [j].DataSet.BookmarkValid (FRanges [j].Bmrk) then
            FRanges [j].DataSet.GotoBookmark (FRanges [j].Bmrk);
        finally
          FRanges [j].DataSet.FreeBookmark (FRanges [j].Bmrk);
          FRanges [j].Bmrk := nil;
          FRanges [j].BmrkOrder:=0;
        end;
      end;
    end;
  end;
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents ('OoReports', [TOdsReport]);
end;

initialization
  {$i ooda.lrs}

end.

